import java.util.*;

public class Elevator {

    private static final int minFloor = 0;
    private static final int maxFloor = 10;
    private static int processingTime = 50;

    private int currentFloor;
    private Direction currentDirection;

    private Map<Integer, List<Integer>> requestedPathsMap;
    // The purpose of currentFloorDestinations is to check which direction the lift must move based
    // on the requests that the current occupants of destinations,
    // the shortest path for current occupants will get the highest priority
    private Boolean[] currentFloorDestinations;

    public Elevator() {
        this.currentFloor = 0;
        this.currentDirection = Direction.UP;
        this.requestedPathsMap = new HashMap<>();
        this.currentFloorDestinations = new Boolean[maxFloor + 1];

        Arrays.fill(this.currentFloorDestinations, Boolean.FALSE);
    }

    public void setProcessingTime(int processingTime) {
        Elevator.processingTime = processingTime;
    }

    public int getCurrentFloor() {
        return this.currentFloor;
    }

    public Map<Integer, List<Integer>> getRequestedPathsMap() {
        return this.requestedPathsMap;
    }

    public Boolean[] getCurrentFloorDestinations() {
        return this.currentFloorDestinations;
    }

    public void start() throws InterruptedException {
        currentDirection = Direction.UP;

        do {
            System.out.println("--------");
            processFloor(currentFloor);
            System.out.println("--------");
        } while(currentDirection != Direction.IDLE);

        System.out.println("No one is waiting and " +
                "no one is looking to go anywhere");
        System.out.println("Chilling for now");
    }

    public void lunchtimeElevatorRush() throws InterruptedException {
        Random random = new Random();
        for (int i = 0; i < 30; i++) {
            callElevator(random.nextInt(11),
                    random.nextInt(10) + 1);
        }
    }

    // implement
    public void callElevator(int start, int destination) throws InterruptedException {
        if (isInvalidFloor(start) && isInvalidFloor(destination)) {
            System.out.println("Both floors entered are invalid: " + start + ", " + destination);
        }
        else if(isInvalidFloor(start) || isInvalidFloor(destination)) {
            if (isInvalidFloor(start)) {
                System.out.println("Starting floor entered is invalid: " + start);
            } else if(isInvalidFloor(destination)) {
                System.out.println("Destination floor entered is invalid: " + destination);
            }
        }
        else{
            if(requestedPathsMap.containsKey(start)) {
                requestedPathsMap.get(start).add(destination);
            }
            else {
                List<Integer> floorRequest = new ArrayList<>();
                floorRequest.add(destination);
                requestedPathsMap.put(start, floorRequest);
            }
        }
    }

    // implement
    private void processFloor(int floor) throws InterruptedException {
        // remove current floor from currentFloorDestinations.
        currentFloorDestinations[floor] = false;
        // add all path requests of current floor.
        if(requestedPathsMap.containsKey(floor)) {
            for(int x : requestedPathsMap.get(floor)) {
                if(!currentFloorDestinations[x]){
                    System.out.println("Boarding at floor: " + floor);
                    currentFloorDestinations[x] = true;
                }
            }
        }

        // remove current floor from all values of requestedPathsMap.
        for (Integer key: requestedPathsMap.keySet()) {
            if(key < floor && currentDirection == Direction.UP) {
                List<Integer> requestedFloors = requestedPathsMap.get(key);
                int toRemove = -1;
                if(requestedFloors.contains(floor)) {
                    for (int i = 0; i < requestedFloors.size(); i++) {
                        if(requestedFloors.get(i) == floor) {
                            toRemove = i;
                            System.out.println("UNBOARDING:" + requestedFloors.get(i));
                        }
                    }
                }
                if(toRemove != -1) {
                    requestedPathsMap.get(key).remove(toRemove);

                }
            }
            else if(key > floor && currentDirection == Direction.DOWN) {
                List<Integer> requestedFloors = requestedPathsMap.get(key);
                int toRemove = -1;
                if(requestedFloors.contains(floor)) {
                    for (int i = 0; i < requestedFloors.size(); i++) {
                        if(requestedFloors.get(i) == floor) {
                            toRemove = i;
                            System.out.println("UNBOARDING:" + requestedFloors.get(i));
                        }
                    }
                }
                if(toRemove != -1) {
                    requestedPathsMap.get(key).remove(toRemove);
                }
            }
        }

        // the lift will continue to move in the direction it is currently moving in
        // unless there are no further requests in that particular direction.
        if(currentDirection == Direction.UP){
            for(int i = floor; i > -1; i--) {
                if (currentFloorDestinations[i]) {
                    currentDirection = Direction.DOWN;
                    break;
                }
            }
            for(int i = floor; i < 11; i++) {
                if (currentFloorDestinations[i]) {
                    currentDirection = Direction.UP;
                    break;
                }
            }
        }
        else if(currentDirection == Direction.DOWN){
            for(int i = floor; i < 11; i++) {
                if (currentFloorDestinations[i]) {
                    currentDirection = Direction.UP;
                    break;
                }
            }
            for(int i = floor; i > -1; i--) {
                if (currentFloorDestinations[i]) {
                    currentDirection = Direction.DOWN;
                    break;
                }
            }
        }
        moveElevator();

    }

    // implement
    private void moveElevator() throws InterruptedException {
        boolean allEmpty = true;
        // check if all requestedPathsMap values are empty.
        // if true set currentDirection to IDLE.
        for (List<Integer> requestedFloors: requestedPathsMap.values()) {
            if (!requestedFloors.isEmpty()) {
                allEmpty = false;
                break;
            }
        }
        if(allEmpty) {
            currentDirection = Direction.IDLE;
        }
        if(currentDirection == Direction.UP){
            moveUp();
        }
        else if(currentDirection == Direction.DOWN) {
            moveDown();
        }
    }

    private void moveUp() throws InterruptedException {
        currentFloor++;
        System.out.println("GOING UP TO " + currentFloor);
        Thread.sleep(processingTime);
    }

    private void moveDown() throws InterruptedException {
        currentFloor--;
        System.out.println("GOING DOWN TO " + currentFloor);
        Thread.sleep(processingTime);
    }

    private boolean isInvalidFloor(int floor) {
        return floor < minFloor || floor > maxFloor;
    }
}
